#########################################################################
# Copyright (c) 2006 DiBcom S.A., All rights reserved
#########################################################################
#
# File        : $RCSfile: dibvue.tcl,v $
# Date        : $Date: 2010/02/11 10:55:28 $
# Version     : $Revision: 1.10 $
# Author      : Stephane ANDRIEUX
# Description : 
# Usage       : 
#
# $Log: dibvue.tcl,v $
# Revision 1.10  2010/02/11 10:55:28  sandrieu
# Added the use of the DIB_HOME env. var. Mandatory.
#
# Revision 1.9  2009/10/12 10:00:33  sandrieu
# Beta. New socket system. Replaced the custom socket by the standard tcl/perl/unix socket.
#
# Revision 1.8  2009/10/02 15:34:40  sandrieu
# Beta. Added a start_socket_vue procedure to re-start the socket when it unexpectedly stops (bug to be fixed).
#
# Revision 1.7  2009/09/29 16:08:54  sandrieu
# Alpha. Improved the socket interface while reading the GDS.
#
# Revision 1.6  2009/09/23 13:27:42  sandrieu
# Alpha. Fixed bug in the socket interface.
#
# Revision 1.5  2009/09/17 14:15:14  sandrieu
# Alpha. Moved the script in the procedure dibvue to be able to pass arguments.
#
# Revision 1.4  2009/09/17 10:56:56  sandrieu
# Alpha. Added global dibvue_line_list to be able to clear all lines.
#
# Revision 1.3  2009/09/17 10:16:27  sandrieu
# Alpha. Improved the socket communication.
#
# Revision 1.2  2009/09/17 09:34:07  sandrieu
# Alpha. Added a socket file name and a flag system to allow SOC Encounter to close the socket. The socket file name uses the encounter pid.
#
# Revision 1.1  2009/09/16 19:00:07  sandrieu
# Under construction. Socket/Bridge between encounter and dibvue.pl .
#
#########################################################################

proc dibvue { {txt_file ""} {gds_file ""} } {
  global env dibvue_socket_id dibvue_port dibvue_line_list
  
  set DESIGN_DIR $env(DESIGN_DIR)
  
  ########################################################
  # 
  # Socket callback procedure
  # 
  ########################################################
  proc dibvue_server {channel clientaddr clientport} {
    global dibvue_line_list
    while {[gets $channel line] >= 0} { 
      eval $line
    }
    close $channel
  }

  ########################################################
  # 
  # Start a socket if not already started before.
  # Automatic port number selection.
  # 
  ########################################################
  if { ![info exist dibvue_socket_id] } {
    set dibvue_socket_id [socket -server dibvue_server 0];
    set dibvue_port [lindex [fconfigure $dibvue_socket_id -sockname] 2]
  }

  if { $txt_file != "" && $gds_file != "" } {
      set cmd_options "-txt $txt_file -gds $gds_file"
  } else { set cmd_options "" }
  
  ########################################################
  # 
  # Start the perl GUI with the socket port number to communicate.
  # 
  ########################################################
  if ![info exists env(DIB_HOME)] { puts "-E- DIB_HOME Environment variable is not set. Please set it and try again. Exit."; 
  } else {eval exec $env(DIB_HOME)/dibvue.pl -socket $dibvue_port $cmd_options & }
  
}

